@@ -3,7 +3,7 @@
  *              that either displays or sets the characteristics of
  *              one or more of the system's networking interfaces.
  *
- * Version:     $Id: ifconfig.c,v 1.33 1999/12/11 13:35:53 freitag Exp $
+ * Version:     $Id: ifconfig.c,v 1.39 2000/06/11 15:44:21 pb Exp $
  *
  * Author:      Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
  *              and others.  Copyright 1993 MicroWalt Corporation
@@ -13,6 +13,10 @@
  *              Public  License as  published  by  the  Free  Software
  *              Foundation;  either  version 2 of the License, or  (at
  *              your option) any later version.
+ *
+ * Patched to support 'add' and 'del' keywords for INET(4) addresses
+ * by Mrs. Brisby <mrs.brisby@nimh.org>
+ *
  * {1.34} - 19980630 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
  *                     - gettext instead of catgets for i18n
  *          10/1998  - Andi Kleen. Use interface list primitives.       
@@ -61,40 +65,8 @@ struct in6_ifreq {
 
 #endif
 
-#define IPV6_ADDR_ANY		0x0000U
-
-#define IPV6_ADDR_UNICAST      	0x0001U
-#define IPV6_ADDR_MULTICAST    	0x0002U
-#define IPV6_ADDR_ANYCAST	0x0004U
-
-#define IPV6_ADDR_LOOPBACK	0x0010U
-#define IPV6_ADDR_LINKLOCAL	0x0020U
-#define IPV6_ADDR_SITELOCAL	0x0040U
-
-#define IPV6_ADDR_COMPATv4	0x0080U
-
-#define IPV6_ADDR_SCOPE_MASK	0x00f0U
-
-#define IPV6_ADDR_MAPPED	0x1000U
-#define IPV6_ADDR_RESERVED	0x2000U		/* reserved address space */
-
 #endif				/* HAVE_AFINET6 */
 
-#ifdef IFF_PORTSEL
-static const char *if_port_text[][4] =
-{
-  /* Keep in step with <linux/netdevice.h> */
-    {"unknown", NULL, NULL, NULL},
-    {"10base2", "bnc", "coax", NULL},
-    {"10baseT", "utp", "tpe", NULL},
-    {"AUI", "thick", "db15", NULL},
-    {"100baseT", NULL, NULL, NULL},
-    {"100baseTX", NULL, NULL, NULL},
-    {"100baseFX", NULL, NULL, NULL},
-    {NULL, NULL, NULL, NULL},
-};
-#endif
-
 #if HAVE_AFIPX
 #if (__GLIBC__ > 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ >= 1)
 #include <netipx/ipx.h>
@@ -110,7 +82,7 @@ static const char *if_port_text[][4] =
 #include "sockets.h"
 #include "util.h"
 
-char *Release = RELEASE, *Version = "ifconfig 1.39 (1999-03-18)";
+char *Release = RELEASE, *Version = "ifconfig 1.40 (2000-05-21)";
 
 int opt_a = 0;			/* show all interfaces          */
 int opt_i = 0;			/* show the statistics          */
@@ -118,247 +90,22 @@ int opt_v = 0;			/* debugging output flag        */
 
 int addr_family = 0;		/* currently selected AF        */
 
+/* for ipv4 add/del modes */
+static int get_nmbc_parent(char *parent,
+			unsigned long *nm, unsigned long *bc);
+static int set_ifstate(char *parent,
+			unsigned long ip,
+			unsigned long nm,
+			unsigned long bc,
+			int flag);
 
-void ife_print(struct interface *ptr)
-{
-    struct aftype *ap;
-    struct hwtype *hw;
-    int hf;
-    int can_compress = 0;
-#if HAVE_AFIPX
-    static struct aftype *ipxtype = NULL;
-#endif
-#if HAVE_AFECONET
-    static struct aftype *ectype = NULL;
-#endif
-#if HAVE_AFATALK
-    static struct aftype *ddptype = NULL;
-#endif
-#if HAVE_AFINET6
-    FILE *f;
-    char addr6[40], devname[20];
-    struct sockaddr_in6 sap;
-    int plen, scope, dad_status, if_idx;
-    extern struct aftype inet6_aftype;
-    char addr6p[8][5];
-#endif
-
-    ap = get_afntype(ptr->addr.sa_family);
-    if (ap == NULL)
-	ap = get_afntype(0);
-
-    hf = ptr->type;
-
-    if (hf == ARPHRD_CSLIP || hf == ARPHRD_CSLIP6)
-	can_compress = 1;
-
-    hw = get_hwntype(hf);
-    if (hw == NULL)
-	hw = get_hwntype(-1);
-
-    printf(_("%-9.9s Link encap:%s  "), ptr->name, hw->title);
-    /* For some hardware types (eg Ash, ATM) we don't print the 
-       hardware address if it's null.  */
-    if (hw->print != NULL && (! (hw_null_address(hw, ptr->hwaddr) &&
-				  hw->suppress_null_addr)))
-	printf(_("HWaddr %s  "), hw->print(ptr->hwaddr));
-#ifdef IFF_PORTSEL
-    if (ptr->flags & IFF_PORTSEL) {
-	printf(_("Media:%s"), if_port_text[ptr->map.port][0]);
-	if (ptr->flags & IFF_AUTOMEDIA)
-	    printf(_("(auto)"));
-    }
-#endif
-    printf("\n");
-
-#if HAVE_AFINET
-    if (ptr->has_ip) {
-	printf(_("          %s addr:%s "), ap->name,
-	       ap->sprint(&ptr->addr, 1));
-	if (ptr->flags & IFF_POINTOPOINT) {
-	    printf(_(" P-t-P:%s "), ap->sprint(&ptr->dstaddr, 1));
-	}
-	if (ptr->flags & IFF_BROADCAST) {
-	    printf(_(" Bcast:%s "), ap->sprint(&ptr->broadaddr, 1));
-	}
-	printf(_(" Mask:%s\n"), ap->sprint(&ptr->netmask, 1));
-    }
-#endif
-
-#if HAVE_AFINET6
-    /* FIXME: should be integrated into interface.c.   */
-
-    if ((f = fopen(_PATH_PROCNET_IFINET6, "r")) != NULL) {
-	while (fscanf(f, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
-		      addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		      addr6p[4], addr6p[5], addr6p[6], addr6p[7],
-		  &if_idx, &plen, &scope, &dad_status, devname) != EOF) {
-	    if (!strcmp(devname, ptr->name)) {
-		sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
-			addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-			addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-		inet6_aftype.input(1, addr6, (struct sockaddr *) &sap);
-		printf(_("          inet6 addr: %s/%d"),
-		 inet6_aftype.sprint((struct sockaddr *) &sap, 1), plen);
-		printf(_(" Scope:"));
-		switch (scope) {
-		case 0:
-		    printf(_("Global"));
-		    break;
-		case IPV6_ADDR_LINKLOCAL:
-		    printf(_("Link"));
-		    break;
-		case IPV6_ADDR_SITELOCAL:
-		    printf(_("Site"));
-		    break;
-		case IPV6_ADDR_COMPATv4:
-		    printf(_("Compat"));
-		    break;
-		case IPV6_ADDR_LOOPBACK:
-		    printf(_("Host"));
-		    break;
-		default:
-		    printf(_("Unknown"));
-		}
-		printf("\n");
-	    }
-	}
-	fclose(f);
-    }
-#endif
-
-#if HAVE_AFIPX
-    if (ipxtype == NULL)
-	ipxtype = get_afntype(AF_IPX);
-
-    if (ipxtype != NULL) {
-	if (ptr->has_ipx_bb)
-	    printf(_("          IPX/Ethernet II addr:%s\n"),
-		   ipxtype->sprint(&ptr->ipxaddr_bb, 1));
-	if (ptr->has_ipx_sn)
-	    printf(_("          IPX/Ethernet SNAP addr:%s\n"),
-		   ipxtype->sprint(&ptr->ipxaddr_sn, 1));
-	if (ptr->has_ipx_e2)
-	    printf(_("          IPX/Ethernet 802.2 addr:%s\n"),
-		   ipxtype->sprint(&ptr->ipxaddr_e2, 1));
-	if (ptr->has_ipx_e3)
-	    printf(_("          IPX/Ethernet 802.3 addr:%s\n"),
-		   ipxtype->sprint(&ptr->ipxaddr_e3, 1));
-    }
-#endif
-
-#if HAVE_AFATALK
-    if (ddptype == NULL)
-	ddptype = get_afntype(AF_APPLETALK);
-    if (ddptype != NULL) {
-	if (ptr->has_ddp)
-	    printf(_("          EtherTalk Phase 2 addr:%s\n"), ddptype->sprint(&ptr->ddpaddr, 1));
-    }
-#endif
-
-#if HAVE_AFECONET
-    if (ectype == NULL)
-	ectype = get_afntype(AF_ECONET);
-    if (ectype != NULL) {
-	if (ptr->has_econet)
-	    printf(_("          econet addr:%s\n"), ectype->sprint(&ptr->ecaddr, 1));
-    }
-#endif
-
-    printf("          ");
-    if (ptr->flags == 0)
-	printf(_("[NO FLAGS] "));
-    if (ptr->flags & IFF_UP)
-	printf(_("UP "));
-    if (ptr->flags & IFF_BROADCAST)
-	printf(_("BROADCAST "));
-    if (ptr->flags & IFF_DEBUG)
-	printf(_("DEBUG "));
-    if (ptr->flags & IFF_LOOPBACK)
-	printf(_("LOOPBACK "));
-    if (ptr->flags & IFF_POINTOPOINT)
-	printf(_("POINTOPOINT "));
-    if (ptr->flags & IFF_NOTRAILERS)
-	printf(_("NOTRAILERS "));
-    if (ptr->flags & IFF_RUNNING)
-	printf(_("RUNNING "));
-    if (ptr->flags & IFF_NOARP)
-	printf(_("NOARP "));
-    if (ptr->flags & IFF_PROMISC)
-	printf(_("PROMISC "));
-    if (ptr->flags & IFF_ALLMULTI)
-	printf(_("ALLMULTI "));
-    if (ptr->flags & IFF_SLAVE)
-	printf(_("SLAVE "));
-    if (ptr->flags & IFF_MASTER)
-	printf(_("MASTER "));
-    if (ptr->flags & IFF_MULTICAST)
-	printf(_("MULTICAST "));
-#ifdef HAVE_DYNAMIC
-    if (ptr->flags & IFF_DYNAMIC)
-	printf(_("DYNAMIC "));
-#endif
-
-    printf(_(" MTU:%d  Metric:%d"),
-	   ptr->mtu, ptr->metric ? ptr->metric : 1);
-#ifdef SIOCSKEEPALIVE
-    if (ptr->outfill || ptr->keepalive)
-	printf(_("  Outfill:%d  Keepalive:%d"),
-	       ptr->outfill, ptr->keepalive);
-#endif
-    printf("\n");
-
-    /* If needed, display the interface statistics. */
-
-    if (ptr->statistics_valid) {
-	/* XXX: statistics are currently only printed for the primary address,
-	 *      not for the aliases, although strictly speaking they're shared
-	 *      by all addresses.
-	 */
-	printf("          ");
-
-	printf(_("RX packets:%lu errors:%lu dropped:%lu overruns:%lu frame:%lu\n"),
-	       ptr->stats.rx_packets, ptr->stats.rx_errors,
-	       ptr->stats.rx_dropped, ptr->stats.rx_fifo_errors,
-	       ptr->stats.rx_frame_errors);
-	if (can_compress)
-	    printf(_("             compressed:%lu\n"), ptr->stats.rx_compressed);
-
-	printf("          ");
-
-	printf(_("TX packets:%lu errors:%lu dropped:%lu overruns:%lu carrier:%lu\n"),
-	       ptr->stats.tx_packets, ptr->stats.tx_errors,
-	       ptr->stats.tx_dropped, ptr->stats.tx_fifo_errors,
-	       ptr->stats.tx_carrier_errors);
-	printf(_("          collisions:%lu "), ptr->stats.collisions);
-	if (can_compress)
-	    printf(_("compressed:%lu "), ptr->stats.tx_compressed);
-	if (ptr->tx_queue_len != -1)
-	    printf(_("txqueuelen:%d "), ptr->tx_queue_len);
-	printf("\n");
-    }
-
-    if ((ptr->map.irq || ptr->map.mem_start || ptr->map.dma ||
-	 ptr->map.base_addr)) {
-	printf("          ");
-	if (ptr->map.irq)
-	    printf(_("Interrupt:%d "), ptr->map.irq);
-	if (ptr->map.base_addr >= 0x100)	/* Only print devices using it for 
-						   I/O maps */
-	    printf(_("Base address:0x%x "), ptr->map.base_addr);
-	if (ptr->map.mem_start) {
-	    printf(_("Memory:%lx-%lx "), ptr->map.mem_start, ptr->map.mem_end);
-	}
-	if (ptr->map.dma)
-	    printf(_("DMA chan:%x "), ptr->map.dma);
-	printf("\n");
-    }
-    printf("\n");
-}
 
 static int if_print(char *ifname)
 {
-    int res; 
+    int res;
+
+    if (ife_short)
+	printf(_("Iface   MTU Met    RX-OK RX-ERR RX-DRP RX-OVR    TX-OK TX-ERR TX-DRP TX-OVR Flg\n"));
 
     if (!ifname) {
 	res = for_all_interfaces(do_if_print, &opt_a);
@@ -373,7 +120,6 @@ static int if_print(char *ifname)
     return res; 
 }
 
-
 /* Set a certain interface flag. */
 static int set_flag(char *ifname, short flag)
 {
@@ -394,41 +140,44 @@ static int set_flag(char *ifname, short flag)
     return (0);
 }
 
-
 /* Clear a certain interface flag. */
 static int clr_flag(char *ifname, short flag)
 {
     struct ifreq ifr;
+    int fd;
+
+    if (strchr(ifname, ':')) {
+        /* This is a v4 alias interface.  Downing it via a socket for
+	   another AF may have bad consequences. */
+        fd = get_socket_for_af(AF_INET);
+	if (fd < 0) {
+	    fprintf(stderr, _("No support for INET on this system.\n"));
+	    return -1;
+	}
+    } else
+        fd = skfd;
 
     strcpy(ifr.ifr_name, ifname);
-    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
+    if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
 	fprintf(stderr, _("%s: unknown interface: %s\n"), 
 		ifname, strerror(errno));
 	return -1;
     }
     strcpy(ifr.ifr_name, ifname);
     ifr.ifr_flags &= ~flag;
-    if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
+    if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
 	perror("SIOCSIFFLAGS");
 	return -1;
     }
     return (0);
 }
 
-
 static void usage(void)
 {
-    fprintf(stderr, _("Usage:\n  ifconfig [-a] [-i] [-v] <interface> [[<AF>] <address>]\n"));
-    /* XXX: it would be useful to have the add/del syntax even without IPv6.
-       the 2.1 interface address lists make this natural */
-#ifdef HAVE_AFINET6
-    fprintf(stderr, _("  [add <address>[/<prefixlen>]]\n"));
-#ifdef SIOCDIFADDR
-    fprintf(stderr, _("  [del <address>[/<prefixlen>]]\n"));
-#endif
-    /* XXX the kernel supports tunneling even without ipv6 */
-#endif
+    fprintf(stderr, _("Usage:\n  ifconfig [-a] [-i] [-v] [-s] <interface> [[<AF>] <address>]\n"));
 #if HAVE_AFINET
+    fprintf(stderr, _("  [add <address>[/<prefixlen>]]\n"));
+    fprintf(stderr, _("  [del <address>[/<prefixlen>]]\n"));
     fprintf(stderr, _("  [[-]broadcast [<address>]]  [[-]pointopoint [<address>]]\n"));
     fprintf(stderr, _("  [netmask <address>]  [dstaddr <address>]  [tunnel <address>]\n"));
 #endif
@@ -479,6 +228,7 @@ static int set_netmask(int skfd, struct ifreq *ifr, struct sockaddr *sa)
 int main(int argc, char **argv)
 {
     struct sockaddr sa;
+    struct sockaddr_in sin;
     char host[128];
     struct aftype *ap;
     struct hwtype *hw;
@@ -493,6 +243,9 @@ int main(int argc, char **argv)
     unsigned long prefix_len;
     char *cp;
 #endif
+#if HAVE_AFINET
+    extern struct aftype inet_aftype;
+#endif
 
 #if I18N
     bindtextdomain("net-tools", "/usr/share/locale");
@@ -512,17 +265,26 @@ int main(int argc, char **argv)
 	if (!strcmp(*argv, "-a"))
 	    opt_a = 1;
 
-	if (!strcmp(*argv, "-v"))
-	    opt_v = 1;
+	else if (!strcmp(*argv, "-s"))
+	    ife_short = 1;
 
-	if (!strcmp(*argv, "-V") || !strcmp(*argv, "-version") ||
+	else if (!strcmp(*argv, "-v"))
+	    opt_v = 1;
+	
+	else if (!strcmp(*argv, "-V") || !strcmp(*argv, "-version") ||
 	    !strcmp(*argv, "--version"))
 	    version();
 
-	if (!strcmp(*argv, "-?") || !strcmp(*argv, "-h") ||
+	else if (!strcmp(*argv, "-?") || !strcmp(*argv, "-h") ||
 	    !strcmp(*argv, "-help") || !strcmp(*argv, "--help"))
 	    usage();
 
+	else {
+	    fprintf(stderr, "ifconfig: unknown option %s\n", argv[0]);
+	    fprintf(stderr, "(Use --help for usage information.\n)");
+	    exit(1);
+	}
+
 	argv++;
 	argc--;
     }
@@ -893,98 +655,188 @@ int main(int argc, char **argv)
 	    spp++;
 	    continue;
 	}
-#if HAVE_AFINET6
+#if HAVE_AFINET || HAVE_AFINET6
 	if (!strcmp(*spp, "add")) {
 	    if (*++spp == NULL)
 		usage();
-	    if ((cp = strchr(*spp, '/'))) {
-		prefix_len = atol(cp + 1);
-		if ((prefix_len < 0) || (prefix_len > 128))
-		    usage();
-		*cp = 0;
-	    } else {
-		prefix_len = 0;
-	    }
-	    safe_strncpy(host, *spp, (sizeof host));
-	    if (inet6_aftype.input(1, host, (struct sockaddr *) &sa6) < 0) {
-		inet6_aftype.herror(host);
-		goterr = 1;
-		spp++;
-		continue;
-	    }
-	    memcpy((char *) &ifr6.ifr6_addr, (char *) &sa6.sin6_addr,
-		   sizeof(struct in6_addr));
+#if HAVE_AFINET6
+	    if (strchr(*spp, ':')) {
+		/* INET6 */
+		if ((cp = strchr(*spp, '/'))) {
+		    prefix_len = atol(cp + 1);
+		    if ((prefix_len < 0) || (prefix_len > 128))
+			usage();
+		    *cp = 0;
+		} else {
+		    prefix_len = 0;
+		}
+		safe_strncpy(host, *spp, (sizeof host));
+		if (inet6_aftype.input(1, host, 
+				       (struct sockaddr *) &sa6) < 0) {
+		    inet6_aftype.herror(host);
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		memcpy((char *) &ifr6.ifr6_addr, (char *) &sa6.sin6_addr,
+		       sizeof(struct in6_addr));
 
-	    fd = get_socket_for_af(AF_INET6);
-	    if (fd < 0) {
-		fprintf(stderr, _("No support for INET6 on this system.\n"));
-		goterr = 1;
+		fd = get_socket_for_af(AF_INET6);
+		if (fd < 0) {
+		    fprintf(stderr, 
+			    _("No support for INET6 on this system.\n"));
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		if (ioctl(fd, SIOGIFINDEX, &ifr) < 0) {
+		    perror("SIOGIFINDEX");
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		ifr6.ifr6_ifindex = ifr.ifr_ifindex;
+		ifr6.ifr6_prefixlen = prefix_len;
+		if (ioctl(fd, SIOCSIFADDR, &ifr6) < 0) {
+		    perror("SIOCSIFADDR");
+		    goterr = 1;
+		}
 		spp++;
 		continue;
 	    }
-	    if (ioctl(fd, SIOGIFINDEX, &ifr) < 0) {
-		perror("SIOGIFINDEX");
-		goterr = 1;
-		spp++;
-		continue;
-	    }
-	    ifr6.ifr6_ifindex = ifr.ifr_ifindex;
-	    ifr6.ifr6_prefixlen = prefix_len;
-	    if (ioctl(fd, SIOCSIFADDR, &ifr6) < 0) {
-		perror("SIOCSIFADDR");
-		goterr = 1;
+#endif
+#ifdef HAVE_AFINET
+	    { /* ipv4 address a.b.c.d */
+		unsigned long ip, nm, bc;
+		safe_strncpy(host, *spp, (sizeof host));
+		if (inet_aftype.input(0, host, (struct sockaddr *)&sin) < 0) {
+		    ap->herror(host);
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		fd = get_socket_for_af(AF_INET);
+		if (fd < 0) {
+		    fprintf(stderr, 
+			    _("No support for INET on this system.\n"));
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+
+		memcpy(&ip, &sin.sin_addr.s_addr, sizeof(unsigned long));
+		
+		if (get_nmbc_parent(ifr.ifr_name, &nm, &bc) < 0) {
+			fprintf(stderr, _("Interface %s not initialized\n"),
+				ifr.ifr_name);
+			goterr = 1;
+			spp++;
+			continue;
+		}
+		set_ifstate(ifr.ifr_name, ip, nm, bc, 1);
+		
 	    }
 	    spp++;
 	    continue;
+#else
+	    fprintf(stderr, _("Bad address.\n"));
+#endif
 	}
+#endif
+
+#if HAVE_AFINET || HAVE_AFINET6
 	if (!strcmp(*spp, "del")) {
 	    if (*++spp == NULL)
 		usage();
-	    if ((cp = strchr(*spp, '/'))) {
-		prefix_len = atol(cp + 1);
-		if ((prefix_len < 0) || (prefix_len > 128))
-		    usage();
-		*cp = 0;
-	    } else {
-		prefix_len = 0;
-	    }
-	    safe_strncpy(host, *spp, (sizeof host));
-	    if (inet6_aftype.input(1, host, (struct sockaddr *) &sa6) < 0) {
-		inet6_aftype.herror(host);
-		goterr = 1;
-		spp++;
-		continue;
-	    }
-	    memcpy((char *) &ifr6.ifr6_addr, (char *) &sa6.sin6_addr,
-		   sizeof(struct in6_addr));
 
-	    fd = get_socket_for_af(AF_INET6);
-	    if (fd < 0) {
-		fprintf(stderr, _("No support for INET6 on this system.\n"));
-		goterr = 1;
-		spp++;
-		continue;
-	    }
-	    if (ioctl(fd, SIOGIFINDEX, &ifr) < 0) {
-		perror("SIOGIFINDEX");
-		goterr = 1;
-		spp++;
-		continue;
-	    }
-	    ifr6.ifr6_ifindex = ifr.ifr_ifindex;
-	    ifr6.ifr6_prefixlen = prefix_len;
 #ifdef SIOCDIFADDR
-	    if (ioctl(fd, SIOCDIFADDR, &ifr6) < 0) {
-		fprintf(stderr, "SIOCDIFADDR: %s\n",
-			strerror(errno));
-		goterr = 1;
+#if HAVE_AFINET6
+	    if (strchr(*spp, ':')) {	/* INET6 */
+		if ((cp = strchr(*spp, '/'))) {
+		    prefix_len = atol(cp + 1);
+		    if ((prefix_len < 0) || (prefix_len > 128))
+			usage();
+		    *cp = 0;
+		} else {
+		    prefix_len = 0;
+		}
+		safe_strncpy(host, *spp, (sizeof host));
+		if (inet6_aftype.input(1, host, 
+				       (struct sockaddr *) &sa6) < 0) {
+		    inet6_aftype.herror(host);
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		memcpy((char *) &ifr6.ifr6_addr, (char *) &sa6.sin6_addr,
+		       sizeof(struct in6_addr));
+		
+		fd = get_socket_for_af(AF_INET6);
+		if (fd < 0) {
+		    fprintf(stderr, 
+			    _("No support for INET6 on this system.\n"));
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		if (ioctl(fd, SIOGIFINDEX, &ifr) < 0) {
+		    perror("SIOGIFINDEX");
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		ifr6.ifr6_ifindex = ifr.ifr_ifindex;
+		ifr6.ifr6_prefixlen = prefix_len;
+		if (ioctl(fd, SIOCDIFADDR, &ifr6) < 0) {
+		    fprintf(stderr, "SIOCDIFADDR: %s\n",
+			    strerror(errno));
+		    goterr = 1;
+		}
+		spp++;
+		continue;
 	    }
+#endif
+#ifdef HAVE_AFINET
+	    {
+		/* ipv4 address a.b.c.d */
+		unsigned long ip, nm, bc;
+		safe_strncpy(host, *spp, (sizeof host));
+		if (inet_aftype.input(0, host, (struct sockaddr *)&sin) < 0) {
+		    ap->herror(host);
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		fd = get_socket_for_af(AF_INET);
+		if (fd < 0) {
+		    fprintf(stderr, _("No support for INET on this system.\n"));
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		
+		memcpy(&ip, &sin.sin_addr.s_addr, sizeof(unsigned long));
+		
+		if (get_nmbc_parent(ifr.ifr_name, &nm, &bc) < 0) {
+		    fprintf(stderr, _("Interface %s not initialized\n"),
+			    ifr.ifr_name);
+		    goterr = 1;
+		    spp++;
+		    continue;
+		}
+		set_ifstate(ifr.ifr_name, ip, nm, bc, 0);
+	    }
+	    spp++;
+	    continue;
+#else
+	    fprintf(stderr, _("Bad address.\n"));
+#endif
 #else
 	    fprintf(stderr, _("Address deletion not supported on this system.\n"));
 #endif
-	    spp++;
-	    continue;
 	}
+#endif
+#if HAVE_AFINET6
 	if (!strcmp(*spp, "tunnel")) {
 	    if (*++spp == NULL)
 		usage();
@@ -1111,3 +963,121 @@ int main(int argc, char **argv)
 
     return (goterr);
 }
+struct ifcmd {
+	int flag;
+	unsigned long addr;
+	char *base;
+	int baselen;
+};
+static unsigned char searcher[256];
+
+static int set_ip_using(const char *name, int c, unsigned long ip)
+{
+	struct ifreq ifr;
+	struct sockaddr_in sin;
+
+	strcpy(ifr.ifr_name, name);
+	memset(&sin, 0, sizeof(struct sockaddr));
+	sin.sin_family = AF_INET;
+	sin.sin_addr.s_addr = ip;
+	memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));
+	if (ioctl(skfd, c, &ifr) < 0)
+		return -1;
+	return 0;
+}
+static int do_ifcmd(struct interface *x, struct ifcmd *ptr)
+{
+	char *z, *e;
+	struct sockaddr_in *sin;
+	int i;
+
+	if (do_if_fetch(x) < 0)
+		return 0;
+	if (strncmp(x->name, ptr->base, ptr->baselen) != 0)
+		return 0; /* skip */
+	z = strchr(x->name, ':');
+	if (!z || !*z)
+		return 0;
+	z++;
+	for (e = z; *e; e++)
+		if (*e == '-') /* deleted */
+			return 0;
+	i = atoi(z);
+	searcher[i] = 1;
+
+	/* copy */
+	sin = (struct sockaddr_in *)&x->dstaddr;
+	if (sin->sin_addr.s_addr != ptr->addr) {
+		return 0;
+	}
+	
+	if (ptr->flag) {
+		/* turn UP */
+		if (set_flag(x->name, IFF_UP | IFF_RUNNING) == -1)
+			return -1;
+	} else {
+		/* turn DOWN */
+		if (clr_flag(x->name, IFF_UP) == -1)
+			return -1;
+	}
+
+	return 1; /* all done! */
+}
+
+
+static int get_nmbc_parent(char *parent,
+			   unsigned long *nm, unsigned long *bc)
+{
+	struct interface *i;
+	struct sockaddr_in *sin;
+
+	i = lookup_interface(parent, 1);
+	if (!i)
+		return -1;
+	if (do_if_fetch(i) < 0)
+		return 0;
+	sin = (struct sockaddr_in *)&i->netmask;
+	memcpy(nm, &sin->sin_addr.s_addr, sizeof(unsigned long));
+	sin = (struct sockaddr_in *)&i->broadaddr;
+	memcpy(bc, &sin->sin_addr.s_addr, sizeof(unsigned long));
+	return 0;
+}
+
+static int set_ifstate(char *parent, unsigned long ip,
+		       unsigned long nm, unsigned long bc,
+		       int flag)
+{
+	char buf[IFNAMSIZ];
+	struct ifcmd pt;
+	int i;
+
+	pt.base = parent;
+	pt.baselen = strlen(parent);
+	pt.addr = ip;
+	pt.flag = flag;
+	memset(searcher, 0, sizeof(searcher));
+	i = for_all_interfaces((int (*)(struct interface *,void *))do_ifcmd, 
+			       &pt);
+	if (i == -1)
+		return -1;
+	if (i == 1)
+		return 0;
+
+	/* add a new interface */
+	for (i = 0; i < 256; i++)
+		if (searcher[i] == 0)
+			break;
+	if (i == 256)
+		return -1; /* FAILURE!!! out of ip addresses */
+
+	sprintf(buf, "%s:%d", parent, i);
+	if (set_ip_using(buf, SIOCSIFADDR, ip) == -1)
+		return -1;
+	if (set_ip_using(buf, SIOCSIFNETMASK, nm) == -1)
+		return -1;
+	if (set_ip_using(buf, SIOCSIFBRDADDR, bc) == -1)
+		return -1;
+	if (set_flag(buf, IFF_BROADCAST) == -1)
+		return -1;
+	return 0;
+}
