@@ -6,7 +6,7 @@
  *              NET-3 Networking Distribution for the LINUX operating
  *              system.
  *
- * Version:     $Id: netstat.c,v 1.33 2000/03/19 12:36:39 philip Exp $
+ * Version:     $Id: netstat.c,v 1.38 2000/05/28 15:19:15 pb Exp $
  *
  * Authors:     Fred Baumgarten, <dc6iq@insu1.etec.uni-karlsruhe.de>
  *              Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
@@ -43,6 +43,7 @@
  *                                      (relly needs to be kernel hooked but 
  *                                      this will do in the meantime)
  *                                      minor header file misplacement tidy up.
+ *980815 {1.xx} Stephane Fillod:       X.25 support
  *980411 {1.34} Arnaldo Carvalho        i18n: catgets -> gnu gettext, substitution
  *                                      of sprintf for snprintf
  *10/1998	Andi Kleen              Use new interface primitives.
@@ -78,6 +79,7 @@
 #include <getopt.h>
 #include <sys/param.h>
 #include <sys/socket.h>
+#include <arpa/inet.h>
 #include <netinet/in.h>
 #include <sys/ioctl.h>
 #include <net/if.h>
@@ -119,7 +121,7 @@ typedef enum {
 #define FEATURE_NETSTAT
 #include "lib/net-features.h"
 
-char *Release = RELEASE, *Version = "netstat 1.38 (1999-04-20)", *Signature = "Fred Baumgarten, Alan Cox, Bernd Eckenfels, Phil Blundell, Tuan Hoang and others";
+char *Release = RELEASE, *Version = "netstat 1.39 (2000-05-21)", *Signature = "Fred Baumgarten, Alan Cox, Bernd Eckenfels, Phil Blundell, Tuan Hoang and others";
 
 
 #define E_READ  -1
@@ -541,8 +543,8 @@ static void igmp_do_one(int lnr, const char *line)
     char mcast_addr[128];
 #if HAVE_AFINET6
     struct sockaddr_in6 mcastaddr;
-    char addr6p[8][5];
-    char addr6[128];
+    char addr6[INET6_ADDRSTRLEN];
+    struct in6_addr in6;
     extern struct aftype inet6_aftype;
 #else
     struct sockaddr_in mcastaddr;
@@ -575,13 +577,10 @@ static void igmp_do_one(int lnr, const char *line)
 	num = sscanf( line, "%d %15s %64[0-9A-Fa-f] %d", &idx, device, mcast_addr, &refcnt );
 	if (num == 4) {
 	    /* Demangle what the kernel gives us */
-	    sscanf(mcast_addr,
-		   "%4s%4s%4s%4s%4s%4s%4s%4s",
-		   addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		   addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-	    snprintf(addr6, sizeof(addr6), "%s:%s:%s:%s:%s:%s:%s:%s",
-		     addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		     addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
+	    sscanf(mcast_addr, "%08X%08X%08X%08X",
+		   &in6.s6_addr32[0], &in6.s6_addr32[1],
+           &in6.s6_addr32[2], &in6.s6_addr32[3]);
+        inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	    inet6_aftype.input(1, addr6, (struct sockaddr *) &mcastaddr);
 	    mcastaddr.sin6_family = AF_INET6;
 	} else {
@@ -639,6 +638,57 @@ static void igmp_do_one(int lnr, const char *line)
     }    /* IPV4 */
 }
 
+#if HAVE_AFX25
+static int x25_info(void)
+{
+       FILE *f=fopen(_PATH_PROCNET_X25, "r");
+       char buffer[256],dev[16];
+       int st,vs,vr,sendq,recvq,lci;
+       static char *x25_state[5]=
+       {
+               "LISTENING",
+               "SABM_SENT",
+               "DISC_SENT",
+               "ESTABLISHED",
+               "RECOVERY"
+       };
+       if(!(f=fopen(_PATH_PROCNET_X25, "r")))
+       {
+               if (errno != ENOENT) {
+                       perror(_PATH_PROCNET_X25);
+                       return(-1);
+               }
+               if (flag_arg || flag_ver)
+                       ESYSNOT("netstat","AF X25");
+               if (flag_arg)
+                       return(1);
+               else
+                       return(0);
+       }
+       printf( _("Active X.25 sockets\n"));
+       /* IMHO, Vr/Vs is not very usefull --SF */
+       printf( _("Dest         Source          Device  LCI  State        Vr/Vs  Send-Q  Recv-Q\n"));
+       fgets(buffer,256,f);
+       while(fgets(buffer,256,f))
+       {
+               buffer[10]=0;
+               buffer[20]=0;
+               sscanf(buffer+22,"%s %d %d %d %d %*d %*d %*d %*d %*d %*d %d %d %*d",
+                       dev,&lci,&st,&vs,&vr,&sendq,&recvq);
+               if (!(flag_all || lci))
+                       continue;
+               printf("%-15s %-15s %-7s %-3d  %-11s  %02d/%02d  %-6d  %-6d\n",
+                       buffer,buffer+11,
+                       dev,
+                       lci,
+                       x25_state[st],
+                       vr,vs,sendq,recvq);
+       }
+       fclose(f);
+       return 0;               
+}
+#endif
+
 static int igmp_info(void)
 {
     INFO_GUTS6(_PATH_PROCNET_IGMP, _PATH_PROCNET_IGMP6, "AF INET (igmp)",
@@ -653,7 +703,8 @@ static void tcp_do_one(int lnr, const char *line)
     struct aftype *ap;
 #if HAVE_AFINET6
     struct sockaddr_in6 localaddr, remaddr;
-    char addr6p[16][3], addr6[128];
+    char addr6[INET6_ADDRSTRLEN];
+    struct in6_addr in6;
     extern struct aftype inet6_aftype;
 #else
     struct sockaddr_in localaddr, remaddr;
@@ -670,29 +721,15 @@ static void tcp_do_one(int lnr, const char *line)
     if (strlen(local_addr) > 8) {
 #if HAVE_AFINET6
 	/* Demangle what the kernel gives us */
-	sscanf(local_addr,
-	       "%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7],
-	       addr6p[8], addr6p[9], addr6p[10], addr6p[11],
-	       addr6p[12], addr6p[13], addr6p[14], addr6p[15]);
-	snprintf(addr6, sizeof(addr6), "%s%s:%s%s:%s%s:%s%s:%s%s:%s%s:%s%s:%s%s",
-		 addr6p[3], addr6p[2], addr6p[1], addr6p[0],
-		 addr6p[7], addr6p[6], addr6p[5], addr6p[4],
-		 addr6p[11], addr6p[10], addr6p[9], addr6p[8],
-		 addr6p[15], addr6p[14], addr6p[13], addr6p[12]);
+	sscanf(local_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+           &in6.s6_addr32[2], &in6.s6_addr32[3]);
+	inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &localaddr);
-	sscanf(rem_addr,
-	       "%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s%2s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7],
-	       addr6p[8], addr6p[9], addr6p[10], addr6p[11],
-	       addr6p[12], addr6p[13], addr6p[14], addr6p[15]);
-	snprintf(addr6, sizeof(addr6), "%s%s:%s%s:%s%s:%s%s:%s%s:%s%s:%s%s:%s%s",
-		 addr6p[3], addr6p[2], addr6p[1], addr6p[0],
-		 addr6p[7], addr6p[6], addr6p[5], addr6p[4],
-		 addr6p[11], addr6p[10], addr6p[9], addr6p[8],
-		 addr6p[15], addr6p[14], addr6p[13], addr6p[12]);
+	sscanf(rem_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+	       &in6.s6_addr32[2], &in6.s6_addr32[3]);
+	inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &remaddr);
 	localaddr.sin6_family = AF_INET6;
 	remaddr.sin6_family = AF_INET6;
@@ -788,8 +825,8 @@ static void udp_do_one(int lnr, const char *line)
     int num, local_port, rem_port, d, state, timer_run, uid, timeout;
 #if HAVE_AFINET6
     struct sockaddr_in6 localaddr, remaddr;
-    char addr6p[8][5];
-    char addr6[128];
+    char addr6[INET6_ADDRSTRLEN];
+    struct in6_addr in6;
     extern struct aftype inet6_aftype;
 #else
     struct sockaddr_in localaddr, remaddr;
@@ -809,19 +846,15 @@ static void udp_do_one(int lnr, const char *line)
 
     if (strlen(local_addr) > 8) {
 #if HAVE_AFINET6
-	sscanf(local_addr, "%4s%4s%4s%4s%4s%4s%4s%4s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-	snprintf(addr6, sizeof(addr6), "%s:%s:%s:%s:%s:%s:%s:%s",
-		 addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		 addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
+	sscanf(local_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+	       &in6.s6_addr32[2], &in6.s6_addr32[3]);
+	inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &localaddr);
-	sscanf(rem_addr, "%4s%4s%4s%4s%4s%4s%4s%4s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-	snprintf(addr6, sizeof(addr6), "%s:%s:%s:%s:%s:%s:%s:%s",
-		 addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		 addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
+	sscanf(rem_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+	       &in6.s6_addr32[2], &in6.s6_addr32[3]);
+	inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &remaddr);
 	localaddr.sin6_family = AF_INET6;
 	remaddr.sin6_family = AF_INET6;
@@ -929,8 +962,8 @@ static void raw_do_one(int lnr, const char *line)
     int num, local_port, rem_port, d, state, timer_run, uid, timeout;
 #if HAVE_AFINET6
     struct sockaddr_in6 localaddr, remaddr;
-    char addr6p[8][5];
-    char addr6[128];
+    char addr6[INET6_ADDRSTRLEN];
+    struct in6_addr in6;
     extern struct aftype inet6_aftype;
 #else
     struct sockaddr_in localaddr, remaddr;
@@ -949,19 +982,15 @@ static void raw_do_one(int lnr, const char *line)
 
     if (strlen(local_addr) > 8) {
 #if HAVE_AFINET6
-	sscanf(local_addr, "%4s%4s%4s%4s%4s%4s%4s%4s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-	snprintf(addr6, sizeof(addr6), "%s:%s:%s:%s:%s:%s:%s:%s",
-		 addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		 addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
+	sscanf(local_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+           &in6.s6_addr32[2], &in6.s6_addr32[3]);
+    inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &localaddr);
-	sscanf(rem_addr, "%4s%4s%4s%4s%4s%4s%4s%4s",
-	       addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-	       addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
-	snprintf(addr6, sizeof(addr6), "%s:%s:%s:%s:%s:%s:%s:%s",
-		 addr6p[0], addr6p[1], addr6p[2], addr6p[3],
-		 addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
+	sscanf(rem_addr, "%08X%08X%08X%08X",
+	       &in6.s6_addr32[0], &in6.s6_addr32[1],
+           &in6.s6_addr32[2], &in6.s6_addr32[3]);
+    inet_ntop(AF_INET6, &in6, addr6, sizeof(addr6));
 	inet6_aftype.input(1, addr6, (struct sockaddr *) &remaddr);
 	localaddr.sin6_family = AF_INET6;
 	remaddr.sin6_family = AF_INET6;
@@ -1395,46 +1424,6 @@ static int ipx_info(void)
 }
 #endif
 
-void ife_print(struct interface *ptr)
-{
-    printf("%-5.5s ", ptr->name);
-    printf("%5d %3d ", ptr->mtu, ptr->metric);
-    /* If needed, display the interface statistics. */
-    if (ptr->statistics_valid) {
-	printf("%8lu %6lu %6lu %6lu ",
-	       ptr->stats.rx_packets, ptr->stats.rx_errors,
-	       ptr->stats.rx_dropped, ptr->stats.rx_fifo_errors);
-	printf("%8lu %6lu %6lu %6lu ",
-	       ptr->stats.tx_packets, ptr->stats.tx_errors,
-	       ptr->stats.tx_dropped, ptr->stats.tx_fifo_errors);
-    } else {
-	printf("%-56s", _("     - no statistics available -"));
-    }
-    if (ptr->flags == 0)
-	printf(_("[NO FLAGS]"));
-    if (ptr->flags & IFF_ALLMULTI)
-	printf("A");
-    if (ptr->flags & IFF_BROADCAST)
-	printf("B");
-    if (ptr->flags & IFF_DEBUG)
-	printf("D");
-    if (ptr->flags & IFF_LOOPBACK)
-	printf("L");
-    if (ptr->flags & IFF_PROMISC)
-	printf("M");
-    if (ptr->flags & IFF_NOTRAILERS)
-	printf("N");
-    if (ptr->flags & IFF_NOARP)
-	printf("O");
-    if (ptr->flags & IFF_POINTOPOINT)
-	printf("P");
-    if (ptr->flags & IFF_RUNNING)
-	printf("R");
-    if (ptr->flags & IFF_UP)
-	printf("U");
-    printf("\n");
-}
-
 static int iface_info(void)
 {
     if (skfd < 0) {
@@ -1444,7 +1433,10 @@ static int iface_info(void)
 	}
 	printf(_("Kernel Interface table\n"));
     }
-    printf(_("Iface   MTU Met    RX-OK RX-ERR RX-DRP RX-OVR    TX-OK TX-ERR TX-DRP TX-OVR Flg\n"));
+    if (flag_exp < 2) {
+	ife_short = 1;
+	printf(_("Iface   MTU Met    RX-OK RX-ERR RX-DRP RX-OVR    TX-OK TX-ERR TX-DRP TX-OVR Flg\n"));
+    }
 
     if (for_all_interfaces(do_if_print, &flag_all) < 0) {
 	perror(_("missing interface information"));
@@ -1512,7 +1504,6 @@ int main
 	{"interfaces", 0, 0, 'i'},
 	{"help", 0, 0, 'h'},
 	{"route", 0, 0, 'r'},
-	{"netlink", 2, 0, 'L'},
 #if HAVE_FW_MASQUERADE
 	{"masquerade", 0, 0, 'M'},
 #endif
@@ -1645,7 +1636,7 @@ int main
         flag_inet = flag_inet6 = 1;
 
     flag_arg = flag_tcp + flag_udp + flag_raw + flag_unx + flag_ipx
-	+ flag_ax25 + flag_netrom + flag_igmp;
+	+ flag_ax25 + flag_netrom + flag_igmp + flag_x25;
 
     if (flag_mas) {
 #if HAVE_FW_MASQUERADE && HAVE_AFINET
@@ -1797,6 +1788,19 @@ int main
 		i = 1;
 		ENOSUPP("netstat", "AF AX25");
 	    }
+#endif
+	}
+	if(!flag_arg || flag_x25) {
+#if HAVE_AFX25
+	    /* FIXME */
+	    i = x25_info();
+	    if (i)
+		return(i);
+#else
+	    if (flag_arg) {
+		i = 1;
+		ENOSUPP("netstat", "AF X25");
+	    }
 #endif
 	}
 	if (!flag_arg || flag_netrom) {
